package checkers.basetype;

import java.io.*;
import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.*;
import java.util.regex.Pattern;

import com.sun.source.tree.CompilationUnitTree;
import com.sun.source.tree.Tree;

import checkers.metaquals.QualifierRoot;
import checkers.metaquals.SubtypeOf;
import checkers.metaquals.TypeQualifiers;
import checkers.source.SourceChecker;
import checkers.types.*;
import checkers.util.GraphAnnotationRelations;

import javax.lang.model.*;
import javax.lang.model.element.AnnotationMirror;
import javax.tools.*;
import javax.annotation.processing.*;

/**
 * An abstract {@link SourceChecker} that provides a simple {@link
 * checkers.source.SourceVisitor} implementation for typical assignment and
 * pseudo-assignment checking of annotated types. Most typechecker plug-ins will
 * want to extend this class.
 * 
 * <p>
 * 
 * Subclasses must define the type qualifiers they support and their subtype
 * hierarchy through {@link BaseTypeChecker#getSupportedTypeQualifiers()} and
 * {@link BaseTypeChecker#createAnnotationRelations()}.  This implementation
 * may infer the type qualifier hierarchy if the subclass is annotated with
 * {@code TypeQualifiers} and the type qualifiers are properly meta-annotated.
 * 
 * @see checkers.metaquals
 */
@SupportedSourceVersion(SourceVersion.RELEASE_7)
@SupportedAnnotationTypes({"*"})
public abstract class BaseTypeChecker extends SourceChecker {

    /** An annotation relations to represent type hierarchy **/
    protected AnnotationRelations annoRelations;
    
    /** instance to compare types according to annoRelations hierarchy **/
    private TypeRelations typeRelations;
    
    /** mapping of error keys to customized error messages **/
    private Properties messages = null;
    
    @Override
    public synchronized void init(ProcessingEnvironment processingEnv) {
        super.init(processingEnv);
        this.annoRelations = this.createAnnotationRelations();
    }
    
    @Override
    protected Properties getMessages() {
        if (this.messages != null)
            return this.messages;
        
        this.messages = new Properties();
        try {
            InputStream base = 
                BaseTypeChecker.class.getResourceAsStream("messages.properties");
            
            if (base == null)
                throw new FileNotFoundException();
                    
            this.messages.load(base);
        } catch (IOException e) {
            this.messager.printMessage(Diagnostic.Kind.WARNING,
                    "error reading the base message file");
        }
                  
        try {
            if (this.getClass() != BaseTypeChecker.class) {
                InputStream custom = 
                    this.getClass().getResourceAsStream("messages.properties");
                if (custom != null) {
                    Properties more = new Properties(this.messages);
                    more.load(custom);
                    this.messages = more;
                }
            }
        } catch (IOException e) {
            this.messager.printMessage(Diagnostic.Kind.WARNING,
                    "error reading the message file for "
                    + this.getClass().getSimpleName());
        }
        return this.messages;
    }

    @Override
    protected BaseTypeVisitor<?, ?> getSourceVisitor(CompilationUnitTree root) {
        
        // Try to reflectively load the visitor.
        final String classToLoad = 
            this.getClass().getName().replace("Checker", "Visitor");
        BaseTypeVisitor<?, ?> result = invokeConstructorFor(classToLoad,
                new Class<?>[] { this.getClass(), CompilationUnitTree.class },
                new Object[] { this, root });
        if (result != null) return result;
        
        // If a visitor couldn't be loaded reflectively, return the default.
        return new BaseTypeVisitor<Void, Void>(this, root);
    }

    /**
     * Tests whether the one annotated type is a subtype of another, with
     * respect to the annotations on these types.
     * 
     * This implementation only annotations in {@link getSupportedTypeQualifiers}.
     * Subclasses may wish to ignore annotations that are not related to the
     * type qualifiers they check.  It follows the subtype rules, as specified
     * in {@link TypesRelations}.
     * 
     * @param sup the parent type
     * @param sub the child type
     * @return true iff {@code sub} is a subtype of {@code sup}
     */
    public boolean isSubtype(AnnotatedTypeMirror sup, AnnotatedTypeMirror sub) {
        return typeRelations.isSubtype(sup, sub);
    }

    /**
     * Tests whether the one annotated type is a valid use of another, with respect
     * to the annotations on these types, without descending into generic or
     * array types (i.e. only performing the validity check on the raw type or 
     * outmost array dimension).
     * 
     * In most cases, the usetype simply needs to be a suptype of elemType.
     * 
     * @param elemType  the type of the class (TypeElement)
     * @param useType   the use of the class (instance type)
     * @return  if the useType is a valid use of elemType
     */
    public boolean isValidUse(AnnotatedTypeMirror elemType, AnnotatedTypeMirror useType) {
        return isSubtype(elemType.getErased(), useType.getErased());
    }

    /**
     * Tests whether the variable accessed is an assignable variable or not,
     * given the current scope
     * 
     * @param varType   the annotated variable type
     * @param variable  the tree access tree
     * @return  true iff variable is assignable in the current scope
     */
    public boolean isAssignable(AnnotatedTypeMirror varType, Tree variable) {
        return true;
    }

    /**
     * If the checker class is annotated with {@link
     * TypeQualifiers}, return an unmodifiable set with the same set
     * of classes as the annotation.  If the class is not so annotated,
     * an empty set is returned.
     * 
     * Subclasses may override this method to return an immutable set
     * of their supported type qualifiers.
     *
     * @return the type qualifiers supported this processor, or an empty
     * set if none
     */
    public Set<Class<? extends Annotation>> getSupportedTypeQualifiers() {
        Class<?> classType = this.getClass();
        TypeQualifiers typeQualifiersAnnotation =
            classType.getAnnotation(TypeQualifiers.class);
        if (typeQualifiersAnnotation == null)
            return Collections.emptySet();
        
        Set<Class<? extends Annotation>> typeQualifiers 
            = new HashSet<Class<? extends Annotation>>();
        for (Class<? extends Annotation> qualifier : typeQualifiersAnnotation.value())
            typeQualifiers.add(qualifier);
        return typeQualifiers;
    }

    /**
     * Returns the type qualifier hierarchy graph to be used by this processor.
     * 
     * The implementation builds the type qualifier hierarchy for the
     * {@link getSupportedTypeQualifier supported types} using the 
     * meta-annotations found in them.  The current implementation returns an
     * instance of {@code GraphAnnotationRelations}.
     * 
     * Subclasses may override this method to express any relationships that
     * cannot be infered using meta-annotations (e.g. due to lack of
     * meta-annotations).
     * 
     * @return an annotation relation tree representing the supported qualifiers
     */
    protected AnnotationRelations createAnnotationRelations() {
        AnnotationFactory annoFactory = new AnnotationFactory(env);

        Set<Class<? extends Annotation>> typeQualifiers = getSupportedTypeQualifiers();
        
        GraphAnnotationRelations annotationRelations = new GraphAnnotationRelations();
        for (Class<?> typeQualifier : typeQualifiers) {
            String typeQualifierName = typeQualifier.getCanonicalName();
            AnnotationMirror typeQualifierAnno = annoFactory.fromName(typeQualifierName);
            if (typeQualifier.getAnnotation(QualifierRoot.class) != null) {
                annotationRelations.setRootAnnotation(typeQualifierAnno);
                break;
            }
        }
        
        // Build the actual graph
        for (Class<?> typeQualifier : typeQualifiers) {
            String typeQualifierName = typeQualifier.getCanonicalName();
            AnnotationMirror typeQualifierAnno = annoFactory.fromName(typeQualifierName);
            // Set subtype relations
            if (typeQualifier.getAnnotation(SubtypeOf.class) == null) {
                annotationRelations.addSubtype(typeQualifierAnno, annotationRelations.getRootAnnotation());
                continue;
            }
            for (Class<?> superAnno : typeQualifier.getAnnotation(SubtypeOf.class).value()) {
                AnnotationMirror superAnnotation = null;
                if (superAnno != Void.class) {
                    String superAnnotationName = superAnno.getCanonicalName();
                    superAnnotation = annoFactory.fromName(superAnnotationName);
                }
                annotationRelations.addSubtype(typeQualifierAnno, superAnnotation);
            }
        }
        return annotationRelations;
    }
    
    /**
     * Update the checker with the currently visited root and
     * get the appropriate factory
     * 
     * Subclasses should consider overriding {@link createFactory}, rather
     * than this.
     * 
     * @param root  the currently visited compilation unit
     * @return  the appropriate type factory 
     */
    @Override
    public final AnnotatedTypeFactory getFactory(CompilationUnitTree root) {
        AnnotatedTypeFactory factory = createFactory(root);
        factory.setAnnotationRelations(annoRelations);
        this.typeRelations = createTypeRelations(factory);
        return factory;
    }

    /**
     * Creates a factory without updating the currently visited root
     * 
     * @param root  the currently visited compilation unit
     * @return  the appropriate type factory 
     */
    public AnnotatedTypeFactory createFactory(CompilationUnitTree root) {
        
        // Try to reflectively load the type factory.
        final String classToLoad = 
            this.getClass().getName().replace("Checker", "AnnotatedTypeFactory");
        
        AnnotatedTypeFactory result = invokeConstructorFor(classToLoad, 
                new Class<?>[] { this.getClass(), CompilationUnitTree.class },
                new Object[] { this, root });
        if (result != null) return result;
        
        return new AnnotatedTypeFactory(this, root);
    }
    
    /**
     * Determines the result of invoking the constructor belonging to the class
     * named by {@code name} having the given parameter types on the given
     * arguments. {@code null} is returned if the class cannot be found, or the
     * constructor does not exist or cannot be invoked on the given arguments,
     * 
     * @param <T> the type to which the constructor belongs
     * @param name the name of the class to which the constructor belongs
     * @param paramTypes the types of the constructor's parameters
     * @param args the arguments on which to invoke the constructor
     * @return the result of the constructor invocation on {@code args}, or
     *         null if the constructor does not exist or could not be invoked
     */
    @SuppressWarnings("unchecked")
    private static <T> T invokeConstructorFor(String name,
            Class<?>[] paramTypes, Object[] args) {
        
        // Load the class.
        Class<T> cls = null; 
        try {
            cls = (Class<T>) Class.forName(name);
        } catch (Exception e) {
            return null;
        }
        
        assert cls != null;
        
        // Invoke the constructor.
        try {
            Constructor<T> ctor = cls.getConstructor(paramTypes);
            return ctor.newInstance(args);
        } catch (NoSuchMethodException e) {
            e.printStackTrace();
        } catch (InstantiationException e) {
            e.printStackTrace();
        } catch (IllegalAccessException e) {
            e.printStackTrace();
        } catch (InvocationTargetException e) {
            e.printStackTrace();
        }
    
        // On failure, return null.
        return null;
    }
    
    /**
     * Creates the type subtyping checker using the current type qualifier 
     * hierarchy.
     * 
     * @param factory   the annotated type factory used currently
     * @return  the type relations class to check type subtyping
     */
    protected TypeRelations createTypeRelations(AnnotatedTypeFactory factory) {
        return new TypeRelations(annoRelations);
    }

    /**
     * Returns a regular expression pattern to specify java classes that are not
     * annotated, and thus whose warnings and should be surpressed.
     * 
     * It returns the pattern specified by the user, through the option
     * {@code checkers.skipClasses}; otherwise it returns a pattern that can
     * match no class.
     * 
     * @return pattern of un-annotated classes that should be skipped
     */
    public Pattern getShouldSkip() {
        String unannotatedClasses = System.getProperty("checkers.skipClasses");
        if (unannotatedClasses != null && !unannotatedClasses.trim().equals(""))
            return Pattern.compile(unannotatedClasses);
        else
            return Pattern.compile("_");
    }
}

