package checkers.flow;

import checkers.source.SourceChecker;
import checkers.types.*;
import checkers.types.AnnotatedTypeMirror.*;
import checkers.util.*;

import java.io.PrintStream;
import java.util.*;

import com.sun.source.tree.*;
import com.sun.source.util.*;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.*;
import javax.lang.model.type.*;

/**
 * Provides a generalized flow-sensitive qualifier inference for the checkers
 * framework.
 *
 * <p>
 *
 * This implementation is based largely on {@code javac}'s dataflow analysis
 * module, which may be found in {@code com.sun.tools.javac.comp.Flow}. Note
 * that, whenever possible, this implementation prefers the use of classes in
 * the public API ({@link BitSet}, the Compiler Tree API) over those in
 * {@code com.sun.tools.javac}.
 *
 * <p>
 *
 * A separate instance of the analysis must be created for each compilation
 * unit.
 */
public class Flow extends TreePathScanner<Void, Void> {

    /** Where to print debugging messages, if non-null. */
    public static PrintStream debug = null;
    
    /** The checker to which this instance belongs. */
    protected final SourceChecker checker;

    /** The processing environment to use. */
    protected final ProcessingEnvironment env;

    /** The file that's being analyzed. */
    protected final CompilationUnitTree root;

    /** The annotation to track. */
    protected final AnnotationMirror annotation;

    /** Utility class for getting source positions. */
    protected final SourcePositions source;

    /** Utility class for determining annotated types. */
    protected final AnnotatedTypeFactory factory;
    
    /** Utility class for operations on annotated types. */
    protected final AnnotatedTypes atypes;

    /** Maps variables to a bit index. */
    protected final List<VariableElement> vars;

    /** Stores the results of the analysis (source position to status bit). */
    protected final Map<Location, Boolean> record;

    /** Tracks the annotated state of each variable during flow. */
    protected BitSet anno;

    /** Tracks the annotated state of each variable in a true branch. */
    protected BitSet annoWhenTrue;

    /** Tracks the annotated state of each variable in a false branch. */
    protected BitSet annoWhenFalse;

    /** Minimal liveness bit, used for better tracking with jumps. */
    private boolean alive = true;

    private boolean tryAlive = true;

    private Stack<Tree> tries = new Stack<Tree>();
    private Stack<BitSet> tryBits = new Stack<BitSet>();

    /** Visitor state; tracking is required for checking receiver types. */
    private final VisitorState visitorState;

    protected static final class Location {
        public final Long position;
        public final Tree.Kind kind;
        public Location(long position, Tree.Kind kind) {
            this.position = position;
            this.kind = kind;
        }
        @Override
        public boolean equals(Object obj) {
            if (this == obj) return true;
            if (!(obj instanceof Location)) return false;
            Location l = (Location)obj;
            return this.position.equals(l.position) && this.kind == l.kind;
        }
        @Override
        public int hashCode() {
            return 47 * position.hashCode() + 113 * kind.hashCode();
        }
        @Override
        public String toString() {
            return "{" + this.position + ", " + this.kind + "}";
        }
    }

    /**
     * Creates a new analysis. The analysis will use the given {@link
     * AnnotatedTypeFactory} to obtain annotated types.
     *
     * @param env the current checker
     * @param root the compilation unit that will be scanned
     * @param annotation the annotation to track
     * @param factory the factory class that will be used to get annotated
     *        types, or {@code null} if the default factory should be used
     */
    public Flow(SourceChecker checker, CompilationUnitTree root,
            AnnotationMirror annotation, AnnotatedTypeFactory factory) {

        this.checker = checker;
        this.env = checker.getProcessingEnvironment();
        this.root = root;
        this.annotation = annotation;

        this.source = Trees.instance(env).getSourcePositions();
        if (factory == null)
            this.factory = new AnnotatedTypeFactory(checker, root);
        else this.factory = factory;
        
        this.atypes = new AnnotatedTypes(env, factory);
        
        this.visitorState = factory.getVisitorState();

        this.vars = new ArrayList<VariableElement>();
        this.record = new HashMap<Location, Boolean>();

        this.anno = new BitSet();
        this.annoWhenTrue = null;
        this.annoWhenFalse = null;
    }

    @Override
    public Void scan(Tree tree, Void p) {
        if (tree != null && getCurrentPath() != null)
            this.visitorState.setPath(new TreePath(getCurrentPath(), tree));
        return super.scan(tree, p);
    }

    /**
     * Tests a result.
     *
     * @param pos the source position of the tree or element to test
     * @return {@link Boolean#TRUE} if the bit for that position is set, {@link
     * Boolean#FALSE} if the bit for that position is not set, or null if there
     * is no result for that position
     */
    private Boolean test(long pos, Tree.Kind kind) {
        Location loc = new Location(pos, kind);
        if (!record.containsKey(loc))
            return null;
        return record.get(loc);
    }
    
    /**
     * Tests a result for a particular tree.
     * 
     * @param tree the tree to test
     * @return {@link Boolean#TRUE} if the bit for that position is set, {@link
     * Boolean#FALSE} if the bit for that position is not set, or null if there
     * is no result for that position
     */
    public Boolean test(Tree tree) {
        if (tree instanceof IdentifierTree || tree instanceof MemberSelectTree) {
            long pos = source.getStartPosition(root, tree);
            return test(pos, tree.getKind());
        }
        return null;
    }

    /**
     * Registers a new variable for flow tracking using its definition.
     *
     * @param tree the variable to register
     */
    void newVar(VariableTree tree) {

        VariableElement var = TreeUtils.elementFromDeclaration(tree);
        assert var != null : "no symbol from tree";

        if (vars.contains(var)) {
            if (debug != null)
                debug.println("Flow: newVar(" + tree + ") reusing index");
            return;
        }

        int idx = vars.size();
        vars.add(var);

        AnnotatedTypeMirror type = factory.getAnnotatedType(tree);
        assert type != null : "no type from symbol";

        if (debug != null)
            debug.println("Flow: newVar(" + tree + ") -- " + type);

        // Determine the initial status of the variable by checking its
        // annotated type.
        if (type.hasAnnotation(annotation))
            anno.set(idx);
        else
            anno.clear(idx);
    }

    /**
     * Moves bits as assignments are made.
     *
     * @param lhs the left-hand side of the assignment
     * @param rhs the right-hand side of the assignment
     */
    void propagate(Tree lhs, Tree rhs) {

        if (debug != null) 
            debug.println("Flow: try propagate from " + rhs + " (" + tryAlive + ")");
        
	if (!tryAlive) return;

	// Skip assignment to arrays.
	if (lhs.getKind() == Tree.Kind.ARRAY_ACCESS)
	    return;
	
        // Get the element for the left-hand side.
        Element elt = InternalUtils.symbol(lhs);
        assert elt != null;

        // Get the element for the right hand side.
        AnnotatedTypeMirror type = factory.getAnnotatedType(rhs);
        assert type != null;

        int idx = vars.indexOf(elt);
        // FIXME: this assertion is valid pending a fix for some field access
        // issues.
        // assert idx >= 0;
        if (idx < 0) return;

        Element rElt = InternalUtils.symbol(rhs);/* TODO type.getElement();*/
        int rIdx = vars.indexOf(rElt);

        if (debug != null)
            debug.println("Flow: (tree) propagate to " + lhs 
                + " from " + type + (rIdx >= 0 ? "/" + anno.get(rIdx) : ""));

        // Propagate/clear the bit if it's annotated or had its bit set.
        if (type.hasAnnotation(this.annotation) || (rIdx >= 0 && anno.get(rIdx)))
            anno.set(idx);
        else anno.clear(idx);
    }
    
    /**
     * Moves bits in an assignment using a type instead of a tree.
     * 
     * @param lhs the left-hand side of the assignment
     * @param rhs the type of the right-hand side of the assignment
     */
    void propagateFromType(Tree lhs, AnnotatedTypeMirror rhs) {
        
        if (!tryAlive) return;
        
        if (lhs.getKind() == Tree.Kind.ARRAY_ACCESS)
            return;
        
        Element elt = InternalUtils.symbol(lhs);
        
        int idx = vars.indexOf(elt);
        if (idx < 0) return;
        
        if (debug != null)
            debug.println("Flow: (type) propagate to " + lhs 
                + " from " + rhs + " (" + rhs.hasAnnotation(this.annotation) + ") "
                + (idx >= 0 ? "/" + anno.get(idx) : ""));
        
        if (rhs.hasAnnotation(this.annotation))
            anno.set(idx);
        else anno.clear(idx);
        
    }

    /**
     * Split the bitset before a conditional branch.
     */
    void split() {
        annoWhenFalse = (BitSet)anno.clone();
        annoWhenTrue = anno;
        anno = null;
    }

    /**
     * Merge the bitset after a conditional branch.
     */
    void merge() {
        anno = (BitSet)annoWhenTrue.clone();
        anno.and(annoWhenFalse);
        //annoWhenTrue = annoWhenFalse = null;
    }

    /**
     * @param path the path to check
     * @return true if the path leaf is part of an expression used as an lvalue
     */
    private boolean isLValue(TreePath path) {
        Tree last = null;
        for (Tree tree : path) {
            if (tree instanceof AssignmentTree)
                return last == ((AssignmentTree)tree).getVariable();
            else if (tree instanceof CompoundAssignmentTree)
                return last == ((CompoundAssignmentTree)tree).getVariable();
//            else if (tree instanceof EnhancedForLoopTree)
//                return last == ((EnhancedForLoopTree)tree).getVariable();
            last = tree;
            break;
        }
        return false;
    }
    
    /**
     * Record the value of the annotation bit for the given usage of a
     * variable, so that a type-checker may use its value after the analysis
     * has finished.
     *
     * @param path
     */
    protected void recordBits(TreePath path) {

        if (isLValue(path))
            return;
        
        Tree tree = path.getLeaf();

        Element elt = null;
        if (tree instanceof MemberSelectTree)
            elt = TreeUtils.elementFromUse((MemberSelectTree)tree);
        else if (tree instanceof IdentifierTree)
            elt = TreeUtils.elementFromUse((IdentifierTree)tree);
        else return;

        int idx = vars.indexOf(elt);

        if (idx >= 0) {
            long pos = source.getStartPosition(root, tree);
            if (debug != null)
                debug.println("Flow: recordBits(" + tree + ") @" + pos + " "
                        + anno.get(idx) + " as " + tree.getKind());
            record.put(new Location(pos, tree.getKind()), anno.get(idx));
        }

    }

    /**
     * Called whenever a definition is scanned.
     *
     * @param tree the definition being scanned
     */
    protected void scanDef(Tree tree) {
        alive = true;
        scan(tree, null);
    }

    /**
     * Called whenever a statement is scanned.
     *
     * @param tree the statement being scanned
     */
    protected void scanStat(StatementTree tree) {
        alive = true;
        scan(tree, null);
    }

    /**
     * Called whenever a block of statements is scanned.
     *
     * @param trees the statements being scanned
     */
    protected void scanStats(List<? extends StatementTree> trees) {
        scan(trees, null);
    }

    /**
     * Called whenever a conditional expression is scanned.
     *
     * @param tree the condition being scanned
     */
    protected void scanCond(Tree tree) {
        alive = true;
        scan(tree, null);
        if (anno != null) split();
        anno = null;
    }

    /**
     * Called whenever an expression is scanned.
     *
     * @param tree the expression being scanned
     */
    protected void scanExpr(ExpressionTree tree) {
        alive = true;
        scan(tree, null);
        if (anno == null) merge();
    }

    /**
     * Called whenever a group of expressions is scanned.
     *
     * @param trees the expressions being scanned
     */
    protected void scanExprs(List<? extends ExpressionTree> trees) {
        for (ExpressionTree tree : trees)
            scanExpr(tree);

    }

    @Override
    public Void visitClass(ClassTree node, Void p) {
        AnnotatedDeclaredType preACT = visitorState.getClassType();
        ClassTree preCT = visitorState.getClassTree();

        visitorState.setClassType((AnnotatedDeclaredType)factory.getAnnotatedType(node));
        visitorState.setClassTree(node);

        try { 
            return super.visitClass(node, p);
        } finally {
            this.visitorState.setClassType(preACT);
            this.visitorState.setClassTree(preCT);
        }
    }

    @Override
    public Void visitImport(ImportTree tree, Void p) {
        return null;
    }

    @Override
    public Void visitAnnotation(AnnotationTree tree, Void p) {
        return null;
    }

    @Override
    public Void visitIdentifier(IdentifierTree node, Void p) {
        super.visitIdentifier(node, p);
        recordBits(getCurrentPath());
        return null;
    }

    @Override
    public Void visitMemberSelect(MemberSelectTree node, Void p) {
        super.visitMemberSelect(node, p);
        recordBits(getCurrentPath());
        return null;
    }

    @Override
    public Void visitVariable(VariableTree node, Void p) {
        newVar(node);
        ExpressionTree init = node.getInitializer();
        if (init != null) {
            scanExpr(init);
            propagate(node, init);
        }
        return null;
    }

    @Override
    public Void visitAssignment(AssignmentTree node, Void p) {
        ExpressionTree var = node.getVariable();
        ExpressionTree expr = node.getExpression();
        if (!(var instanceof IdentifierTree))
            scanExpr(var);
        scanExpr(expr);
        propagate(var, expr);
        return null;
    }
    
    @Override
    public Void visitEnhancedForLoop(EnhancedForLoopTree node, Void p) {
        VariableTree var = node.getVariable();
        newVar(var);
        
        ExpressionTree expr = node.getExpression();
        scanExpr(expr);
        
        AnnotatedTypeMirror rhs = factory.getAnnotatedType(expr);
        AnnotatedTypeMirror iter = atypes.getIteratedType(rhs);
        if (iter != null)
            propagateFromType(var, iter);
        
        return super.visitEnhancedForLoop(node, p);
    }

    @Override
    public Void visitAssert(AssertTree node, Void p) {
        scanCond(node.getCondition());
        anno = (BitSet)annoWhenTrue.clone();
        return null;
    }

    @Override
    public Void visitIf(IfTree node, Void p) {
        scanCond(node.getCondition());

        BitSet before = annoWhenFalse;
        anno = annoWhenTrue;

        boolean aliveBefore = alive;

        scanStat(node.getThenStatement());
        StatementTree elseStmt = node.getElseStatement();
        if (elseStmt != null) {
            boolean aliveAfter = alive;
            alive = aliveBefore;
            BitSet after = (BitSet)anno.clone();
            anno = before;
            scanStat(elseStmt);
            alive &= aliveAfter;
            if (!alive) {
                anno.clear();
                anno.or(after);
            } else
                anno.and(after);
        } else {
            alive &= aliveBefore;
            if (!alive) {
                anno.clear();
                anno.or(before);
            } else
                anno.and(before);
        }

        return null;
    }
    
    @Override
    public Void visitConditionalExpression(ConditionalExpressionTree node,
            Void p) {
        
        // Split and merge as for an if/else.
        scanCond(node.getCondition());
        
        BitSet before = annoWhenFalse;
        anno = annoWhenTrue;

        scanExpr(node.getTrueExpression());
        BitSet after = (BitSet)anno.clone();
        anno = before;
        
        scanExpr(node.getFalseExpression());
        anno.and(after);
        
        return null;
    }

    @Override
    public Void visitWhileLoop(WhileLoopTree node, Void p) {
        boolean pass = false;
        BitSet annoCond;
        do {
            BitSet annoEntry = (BitSet)anno.clone();
            scanCond(node.getCondition());
            annoCond = annoWhenFalse;
            anno = annoWhenTrue;
            scanStat(node.getStatement());
            if (pass) break;
            annoWhenTrue.and(annoEntry);
            pass = true;
        } while (true);
        anno = annoCond;
        return null;
    }
    
    @Override
    public Void visitBreak(BreakTree node, Void p) {
        alive = false;
        return null;
    }

    @Override
    public Void visitContinue(ContinueTree node, Void p) {
        alive = false;
        return null;
    }

    @Override
    public Void visitReturn(ReturnTree node, Void p) {
        if (node.getExpression() != null)
            scanExpr(node.getExpression());
        alive = false;
        return null;
    }

    @Override
    public Void visitThrow(ThrowTree node, Void p) {
        scanExpr(node.getExpression());
        alive = false;
        return null;
    }

    @Override
    public Void visitTry(TryTree node, Void p) {
	tries.push(node);
	tryBits.push((BitSet)anno.clone());
	scan(node.getBlock(), p);
	tries.pop();
	BitSet result = tryBits.pop();
	anno.and(result);
	scan(node.getCatches(), p);
//	if (!alive) {
//	    anno.clear();
//	    anno.or(result);
//	}
	scan(node.getFinallyBlock(), p);
	return null;
    }

    // TODO: visit class to ensure visiting fields before methods
    // TODO: should really check for "java.lang.System", not just "System"
    @Override
    public Void visitMethodInvocation(MethodInvocationTree node, Void p) {
        super.visitMethodInvocation(node, p);

        ExecutableElement method = TreeUtils.elementFromUse(node);
        if (method.getSimpleName().contentEquals("exit")
                && method.getEnclosingElement().getSimpleName().contentEquals("System"))
            alive = false;

	List<? extends TypeMirror> thrown = method.getThrownTypes();
	if (!thrown.isEmpty() && TreeUtils.enclosingOfKind(getCurrentPath(), Tree.Kind.TRY) != null) {
	    if (!tryBits.isEmpty())
	        tryBits.peek().and(anno);
	}

        return null;
    }

    @Override
    public Void visitMethod(MethodTree node, Void p) {
        AnnotatedDeclaredType preMRT = visitorState.getMethodReceiver();
        MethodTree preMT = visitorState.getMethodTree();
        visitorState.setMethodReceiver(
                ((AnnotatedExecutableType)factory.getAnnotatedType(node)).getReceiverType());
        visitorState.setMethodTree(node);
        
        try {
            // Intraprocedural, so save and restore bits.
            BitSet prev = (BitSet)anno.clone();
            super.visitMethod(node, p);
            anno = prev;
            return null;
        } finally {
            visitorState.setMethodReceiver(preMRT);
            visitorState.setMethodTree(preMT);
        }
    }
}
