package checkers.interned;

import java.util.*;

import checkers.source.*;
import checkers.basetype.*;
import checkers.types.*;
import checkers.util.*;
import com.sun.source.tree.*;
import com.sun.source.util.*;

import javax.lang.model.element.*;
import javax.lang.model.type.*;
import static javax.lang.model.util.ElementFilter.*;

/**
 * A type-checking visitor for the {@link checkers.quals.Interned} type
 * qualifier that uses the {@link BaseTypeVisitor} implementation. This visitor
 * reports errors or warnings for violations for the following cases:
 *
 * <ul>
 *  <li>if both sides of a "==" or "!=" comparison are not Interned (error
 *  "not.interned")</li>
 *  <li>if the receiver and argument for a call to an equals method are both
 *  Interned (optional warning "unnecessary.equals")</li>
 * </ul>
 *
 *
 * @see BaseTypeVisitor
 */
public class InternedVisitor extends BaseTypeVisitor<Void, Void> {

    /** The interned annotation. */
    private final AnnotationMirror INTERNED;
    private final AnnotatedTypeMirror INTERNED_OBJECT;

    /**
     * Creates a new visitor for type-checking {@link checkers.quals.Interned}.
     *
     * @param checker the checker to use
     * @param root the root of the input program's AST to check
     */
    public InternedVisitor(InternedChecker checker, CompilationUnitTree root) {
        super(checker, root);
        this.INTERNED = annoFactory.fromName("checkers.quals.Interned");
        INTERNED_OBJECT = AnnotatedTypeMirror.createType(types.getDeclaredType(
                    elements.getTypeElement("java.lang.Object")),
                    checker.getProcessingEnvironment(), factory);
        INTERNED_OBJECT.addAnnotation(INTERNED);
    }

    @Override
    public Void visitBinary(BinaryTree node, Void p) {

        // No checking unless the operator is "==" or "!=".
        if (!(node.getKind() == Tree.Kind.EQUAL_TO ||
                node.getKind() == Tree.Kind.NOT_EQUAL_TO))
            return super.visitBinary(node, p);

        Tree leftOp = node.getLeftOperand(), rightOp = node.getRightOperand();

        // Check passes if one arg is null.
        if (leftOp.getKind() == Tree.Kind.NULL_LITERAL ||
                rightOp.getKind() == Tree.Kind.NULL_LITERAL)
            return super.visitBinary(node, p);

        // Heuristically check that the comparison is the member of a class of
        // comparisons that should be skipped.
        if (suppressByHeuristic(node))
            return super.visitBinary(node, p);

        AnnotatedTypeMirror left = factory.getAnnotatedType(leftOp);
        AnnotatedTypeMirror right = factory.getAnnotatedType(rightOp);

        // Check passes due to auto-unboxing.
        if (left.getKind().isPrimitive() || right.getKind().isPrimitive())
            return super.visitBinary(node, p);

        if (!checker.isSubtype(INTERNED_OBJECT, left))
            checker.report(Result.failure("not.interned", left), leftOp);
        if (!checker.isSubtype(INTERNED_OBJECT, right))
            checker.report(Result.failure("not.interned", right), rightOp);

        return super.visitBinary(node, p);
    }

    @Override
    public Void visitMethodInvocation(MethodInvocationTree node, Void p) {
        if (isInvocationOfEquals(node)) {
            AnnotatedTypeMirror recv = factory.getReceiver(node);
            AnnotatedTypeMirror comp = factory.getAnnotatedType(node.getArguments().get(0));

            if (this.checker.getLintOption("dotequals", true)
                    && checker.isSubtype(INTERNED_OBJECT, recv)
                    && checker.isSubtype(INTERNED_OBJECT, comp))
                checker.report(Result.warning("unnecessary.equals"), node);
        }

        return super.visitMethodInvocation(node, p);
    }

    /**
     * Tests whether a method invocation is an invocation of
     * {@link Object#equals}.
     * 
     * @param node a method invocation node
     * @return true iff {@code node} is a invocation of {@code equals()}
     */
    private boolean isInvocationOfEquals(MethodInvocationTree node) {
        ExecutableElement method = TreeUtils.elementFromUse(node);
        return (method.getParameters().size() == 1
                && method.getReturnType().getKind() == TypeKind.BOOLEAN
                && method.getSimpleName().contentEquals("equals"));
    }

    /**
     * Heuristically determines whether checking for a particular comparison
     * should be suppressed. Specifically, this method tests the following:
     *
     * <ul>
     * <li>the comparison is a == comparison, and</li>
     *
     * <li>it is the test of an if statement that's the first statement in the method,
     * and</li>
     *
     * <li>one of the following is true:
     *
     * <ul>
     * <li>the method overrides {@link Comparator#compare}, the "then" branch
     * of the if statement returns zero, and the comparison tests equality of
     * the method's two parameters</li>
     *
     * <li>the method overrides {@link Object#equals(Object)} and the
     * comparison tests "this" against the method's parameter</li>
     * </ul>
     *
     * </li>
     * </ul>
     *
     * @param node the comparison to check
     * @return true if one of the supported heuristics is matched, false
     *         otherwise
     */
    private boolean suppressByHeuristic(final BinaryTree node) {

        // Only valid if called on an == comparison.
        if (node.getKind() != Tree.Kind.EQUAL_TO)
            return false;

        Tree left = node.getLeftOperand();
        Tree right = node.getRightOperand();

        // Only valid if we're comparing identifiers.
        if (!(left.getKind() == Tree.Kind.IDENTIFIER
                && right.getKind() == Tree.Kind.IDENTIFIER))
            return false;

        // If we're not directly in an if statement in a method (ignoring
        // parens and blocks), terminate.
        if (!Heuristics.matchParents(getCurrentPath(), Tree.Kind.IF, Tree.Kind.METHOD))
            return false;

        // Determine whether or not the "then" statement of the if has a single
        // "return 0" statement (for the Comparator.compare heuristic).
        final boolean returnsZero =
            Heuristics.applyAt(getCurrentPath(), Tree.Kind.IF, new Heuristics.Matcher() {

                @Override
                public Boolean visitIf(IfTree tree, Void p) {
                    return visit(tree.getThenStatement(), p);
                }

                @Override
                public Boolean visitBlock(BlockTree tree, Void p) {
                    if (tree.getStatements().size() > 0)
                        return visit(tree.getStatements().get(0), p);
                    return false;
                }

                @Override
                public Boolean visitReturn(ReturnTree tree, Void p) {
                    ExpressionTree expr = tree.getExpression();
                    return (expr.getKind() == Tree.Kind.INT_LITERAL &&
                            ((LiteralTree)expr).getValue().equals(0));
                }
            });


        ExecutableElement enclosing =
            TreeUtils.elementFromDeclaration(visitorState.getMethodTree());
        assert enclosing != null;

        Element lhs = TreeUtils.elementFromUse((IdentifierTree)left);
        Element rhs = TreeUtils.elementFromUse((IdentifierTree)right);

        if (returnsZero && overrides(enclosing, "java.util.Comparator", "compare")) {
            assert enclosing.getParameters().size() == 2;
            Element p1 = enclosing.getParameters().get(0);
            Element p2 = enclosing.getParameters().get(1);
            return (p1.equals(lhs) && p2.equals(rhs))
                || (p2.equals(lhs) && p1.equals(rhs));

        } else if (overrides(enclosing, "java.lang.Object", "equals")) {
            assert enclosing.getParameters().size() == 1;
            Element param = enclosing.getParameters().get(0);
            Element thisElt = getThis(this.getCurrentPath());
            assert thisElt != null;
            return (thisElt.equals(lhs) && param.equals(rhs))
                || (param.equals(lhs) && thisElt.equals(rhs));
        }

        return false;
    }

    /**
     * Determines the element corresponding to "this" inside a scope.
     *
     * @param path the path to a tree inside the desired scope
     * @return the element corresponding to "this" in the scope of the tree
     *         given by {@code path}
     */
    private final Element getThis(TreePath path) {
        for (Element e : trees.getScope(path).getLocalElements())
            if (e.getSimpleName().contentEquals("this"))
                return e;
        return null;
    }

    /**
     * Determines whether or not the given element overrides the named method in
     * the named class.
     *
     * @param e an element for a method
     * @param clazz the name of a class
     * @param method the name of a method
     * @return true if the method given by {@code e} overrides the named method
     *         in the named class; false otherwise
     */
    private final boolean overrides(
            ExecutableElement e, String clazz, String method) {

        // Get the element named by "clazz".
        TypeElement comp = elements.getTypeElement(clazz);
        if (comp == null) return false;

        // Check all of the methods in the class for name matches and overriding.
        for (ExecutableElement elt : methodsIn(comp.getEnclosedElements()))
            if (elt.getSimpleName().contentEquals(method)
                    && elements.overrides(e, elt, comp))
                return true;

        return false;
    }
}
