package checkers.nonnull;

import java.util.*;

import javax.lang.model.element.*;
import static javax.lang.model.element.Modifier.*;
import javax.lang.model.type.*;

import checkers.basetype.BaseTypeChecker;
import checkers.flow.*;
import checkers.quals.*;
import checkers.types.*;
import checkers.types.AnnotatedTypeMirror.*;
import checkers.util.TreeUtils;
import checkers.util.TypesUtils;

import com.sun.source.tree.*;
import com.sun.source.util.*;

/**
 * An {@link AnnotatedTypeFactory} that accounts for the properties of the
 * NonNull type system. This type factory will add the {@link NonNull}
 * annotation to a type if the input:
 *
 * <ul>
 * <li>is determined to be NonNull by flow-sensitive inference</li>
 * <li>is the class in a static member access (e.g., "System" in
 * "System.out")</li>
 * <li>is an array-creation expression (with new)</li>
 * <li>is an object-creation expression (with new)</li>
 * <li>is a string literal</li>
 * <li>is a package declaration</li>
 * </ul>
 *
 * TODO: some are missing from the above list
 *
 * <p>
 *
 * Additionally, the type factory will add the {@link Nullable} annotation to a
 * type if the input is the null literal.
 */
public class NonNullAnnotatedTypeFactory extends AnnotatedTypeFactory {

    private final Flow flow;
    private final TypeAnnotator typeAnnotator;
    private final TreeAnnotator treeAnnotator;
    private final SourcePositions srcPos;
    private final AnnotationCompleter completer;
    
    /** Represents {@link NonNull}. */
    final AnnotationMirror NONNULL;
    
    /** Represents {@link Raw}. */
    final AnnotationMirror RAW;
    
    /** Represents {@link Nullable}. */
    final AnnotationMirror NULLABLE;
    
    /** Creates a {@link NonNullAnnotatedTypeFactory}. */
    public NonNullAnnotatedTypeFactory(NonNullChecker checker,
            CompilationUnitTree root) {
        super(checker, root);
        
        typeAnnotator = new NonNullTypeAnnotator(checker);
        treeAnnotator = new NonNullTreeAnnotator(checker);
        
        srcPos = trees.getSourcePositions();
        
        NONNULL = this.annotations.fromName("checkers.quals.NonNull");
        RAW = this.annotations.fromName("checkers.quals.Raw");
        NULLABLE = this.annotations.fromName("checkers.quals.Nullable");
        
        completer = new AnnotationCompleter(NULLABLE);
        
        flow = new NonNullFlow(checker, root, NONNULL, this);
        //flow.debug = System.out;
        flow.scan(root, null);
    }

    private void checkRep(AnnotatedTypeMirror type, Object src) {
        String info = "";
        if (src instanceof Tree) {
            Tree t = (Tree)src;
            long pos = srcPos.getStartPosition(root, t);
            if (pos == -1)
                info = "(invalid position) " + t.getKind();
            else {
                LineMap lm = root.getLineMap();
                info = String.format("checkRep failure @ %s:%d:%d", 
                        root.getSourceFile().getName(),
                        lm.getLineNumber(pos),
                        lm.getColumnNumber(pos));
            }
        }
        assert !(type.hasAnnotation(NONNULL) && type.hasAnnotation(NULLABLE)) :
            type + ": " + info;
    }
    
    @Override
    protected void annotateImplicit(Element elt, AnnotatedTypeMirror type) {

        if (elt instanceof VariableElement)
            annotateIfStatic(elt, type);
        
        typeAnnotator.visit(type);
        if (!(elt instanceof TypeElement))
            applyDefaults(elt, type);
        completer.visit(type);
        checkRep(type, elt);
    }
    
    @Override
    protected void annotateImplicit(Tree tree, AnnotatedTypeMirror type) {
        treeAnnotator.visit(tree, type);
        typeAnnotator.visit(type);
        annotateDefaults(tree, type);
        boolean rawSub = substituteRaw(tree, type);
        if (flow.test(tree) == Boolean.TRUE) {
            if (!(type.hasAnnotation(RAW) && rawSub)) {
                type.clearAnnotations();
                type.addAnnotation(NONNULL);
            }
        }
        completer.visit(type);
        checkRep(type, tree);
    }
    
    /**
     * Substitutes {@link Raw} annotations on a type. If the select expression
     * (which might implicitly be "this") is {@link Raw}, this replaces the
     * annotations on {@code type} with {@link Raw}.
     * 
     * @param tree a (possibly implicit) select expression
     * @param type the type for which to substitute annotations based on the
     *        placement of {@link Raw}
     * @return true iff we reduced the annotated type from {@link NonNull} to
     *         {@link Raw}
     */
    boolean substituteRaw(Tree tree, AnnotatedTypeMirror type) {
        
        // If it's not an expression tree, it's definitely not a select.
        if (!(tree instanceof ExpressionTree))
            return false;
        
        // Identifiers might implicitly select from "this", but not if they're
        // class or interface identifiers.
        if (tree instanceof IdentifierTree) {
            
            Element elt = TreeUtils.elementFromUse((IdentifierTree) tree);
            if (elt == null || elt.getKind() != ElementKind.FIELD)
                return false;
            
	    Tree decl = declarationFromElement(elt);
	    if (this.getPath(decl) != null 
                    && decl instanceof VariableTree
                    && ((VariableTree) decl).getInitializer() != null 
                    && getAnnotatedType(((VariableTree) decl).getInitializer()).hasAnnotation(NONNULL))
		return false;
	}
        
        final AnnotatedTypeMirror select = getReceiver((ExpressionTree) tree);
        if (select != null && select.hasAnnotation(RAW) && !type.hasAnnotation(NULLABLE)) {
            boolean wasNN = type.hasAnnotation(NONNULL);
            type.clearAnnotations();
            type.addAnnotation(RAW);
            return wasNN;            
        }
        return false;
    }
    
    private Element nearestEnclosing(Tree tree) {
        
        TreePath path = this.getPath(tree);
        assert path != null : tree;
        
        for (Tree t : path) {
            switch (t.getKind()) {
            case VARIABLE:
                return TreeUtils.elementFromDeclaration((VariableTree)t);
            case METHOD:
                return TreeUtils.elementFromDeclaration((MethodTree)t);
            case CLASS:
                return TreeUtils.elementFromDeclaration((ClassTree)t);
            default: // Do nothing.
            }
        }
        
        return null;
    }

    private void annotateDefaults(Tree tree, AnnotatedTypeMirror type) {

        // TODO handling for method invocation type arguments
        
        Element elt;
        switch (tree.getKind()) {
            case MEMBER_SELECT:
                elt = TreeUtils.elementFromUse((MemberSelectTree)tree);
                applyDefaults(elt, type);
                break;
            
            case IDENTIFIER:
                elt = TreeUtils.elementFromUse((IdentifierTree)tree);
                applyDefaults(elt, type);
                break;
                
            case METHOD_INVOCATION:
                elt = TreeUtils.elementFromUse((MethodInvocationTree)tree);
                applyDefaults(elt, type);
                break;
 
            default:
                Element nearest = nearestEnclosing(tree);
                applyDefaults(nearest, type);
        }
    }
    
    private void applyDefaults(final Element elt, final AnnotatedTypeMirror type) {
        assert elt != null : "null element";
        Element e = elt;
        while (e != null) {
            Default d = e.getAnnotation(Default.class);
            if (d != null && d.value().equals(NonNull.class.getName())) {

                // Get the default locations from the Default annotation.
                final Set<DefaultLocation> locations = 
                    new HashSet<DefaultLocation>(Arrays.asList(d.types()));

                new AnnotatedTypeScanner<Void, AnnotationMirror>() {

                    @Override
                    public Void scan(AnnotatedTypeMirror t,
                            AnnotationMirror p) {

                        if (t == null 
                                || t.getKind() == TypeKind.NONE 
                                || t.getKind() == TypeKind.WILDCARD)
                            return null;
                        
                        // Skip type variables.
                        if (t.getKind() == TypeKind.TYPEVAR)
                            return super.scan(t, p);

                        // Skip annotating this type if:
                        // - the default is "all except (the raw types of) locals"
                        // - we are applying defaults to a local
                        // - and the type is a raw type
                        if (elt.getKind() == ElementKind.LOCAL_VARIABLE
                                && locations.contains(DefaultLocation.ALL_EXCEPT_LOCALS)
                                && t == type)
                            return super.scan(t, p);
                         
                        if (!t.hasAnnotation(NULLABLE))
                            t.addAnnotation(p);
                        return super.scan(t, p);
                    }

                    // Skip method receivers.
                    @Override
                    public Void visitExecutable(AnnotatedExecutableType t, 
                            AnnotationMirror p) {
                        scan(t.getReturnType(), p);
                        scanAndReduce(t.getParameterTypes(), p, null);
                        scanAndReduce(t.getThrownTypes(), p, null);
                        scanAndReduce(t.getTypeVariables(), p, null);
                        return null;
                    }

                }.scan(type, NONNULL);
            }
	    e = e.getEnclosingElement();
        }
    }
    
    private class AnnotationCompleter extends AnnotatedTypeScanner<Void, Void> {
        
        private final AnnotationMirror annotation;

        /** Creates an {@link AnnotationCompleter} for the given annotation. */
        AnnotationCompleter(AnnotationMirror annotation) {
            this.annotation = annotation;
        }
        
        @Override 
        protected Void scan(AnnotatedTypeMirror type, Void p) {

            if (type == null)
                return super.scan(type, p);

            if (type.getAnnotations().isEmpty())
                type.addAnnotation(annotation);
            else if (type.hasAnnotation(RAW))
                type.removeAnnotation(NONNULL);
            else if (type.hasAnnotation(NONNULL))
                type.removeAnnotation(NULLABLE);
            
            assert !type.getAnnotations().isEmpty() : type;

            return super.scan(type, p);
        }
    }
    
    private class NonNullTypeAnnotator extends TypeAnnotator {

        /** Creates a {@link NonNullTypeAnnotator} for the given checker. */
        NonNullTypeAnnotator(BaseTypeChecker checker) {
            super(checker);
        }

        @Override
        public Void visitExecutable(AnnotatedExecutableType type, Void p) {

            // Don't add implicit receiver annotations if some are already there.
            if (!type.getReceiverType().getAnnotations().isEmpty())
                return super.visitExecutable(type, p);
            
            ExecutableElement elt = type.getElement();
            assert elt != null;

            // Add @Raw to constructors. 
            if (elt.getKind() == ElementKind.CONSTRUCTOR)
                type.getReceiverType().addAnnotation(RAW);
            
            // Add @NonNull to non-static non-constructors.
            if (!elt.getModifiers().contains(STATIC))
                type.getReceiverType().addAnnotation(NONNULL);

            return super.visitExecutable(type, p);
        }
    }
    
    private void annotateIfStatic(
            Element elt, AnnotatedTypeMirror type) {
        
        if (elt == null) 
            return;
        
        if (elt.getKind().isClass() || elt.getKind().isInterface()) {
            type.clearAnnotations();
            type.addAnnotation(NONNULL);
        }
        
        // Workaround for System.{out,in,err} issue: assume all static
        // fields in java.lang.System are nonnull.
        if (elt.getKind().isField()) {
            
            Set<Modifier> mods = elt.getModifiers();
            if (mods.contains(STATIC) && mods.contains(FINAL)) {
                VariableElement var = (VariableElement)elt;
                Element parent = var.getEnclosingElement();
                
                // Heuristic: if we have a static final field in a system package,
                // treat it as NonNull (many like Boolean.TYPE and System.out
                // have constant value null but are set by the VM).
                boolean inJavaPackage = parent != null 
                        && parent instanceof TypeElement 
                        && ((TypeElement)parent).getQualifiedName().toString().startsWith("java.");
                
                if (var.getConstantValue() != null
                        || var.getSimpleName().contentEquals("class")
                        || inJavaPackage) {
                    type.clearAnnotations();
                    type.addAnnotation(NONNULL);
                }
            }
        }
    }
    
    private class NonNullTreeAnnotator extends TreeAnnotator {
        
        /** Creates a {@link NonNullTreeAnnotator} for the given checker. */
        NonNullTreeAnnotator(BaseTypeChecker checker) {
            super(checker);
        }

        @Override
        public Void visitMemberSelect(MemberSelectTree node, AnnotatedTypeMirror type) {
            
            Element elt = TreeUtils.elementFromUse(node);
            assert elt != null;
            
            annotateIfStatic(elt, type);
            return super.visitMemberSelect(node, type);
        }

        @Override
        public Void visitIdentifier(IdentifierTree node, AnnotatedTypeMirror type) {
            
            Element elt = TreeUtils.elementFromUse(node);
            assert elt != null;
            
            if (elt.getKind() == ElementKind.PARAMETER
                    && TypesUtils.isThrowable(elt.asType())) {
                // Workaround: exception parameters should be implicitly
                // NonNull, but due to a compiler bug they have
                // kind PARAMETER instead of EXCEPTION_PARAMETER. Until it's
                // fixed we manually inspect enclosing catch blocks.
                final TreePath path = getPath(node);
                Tree ct = TreeUtils.enclosingOfKind(path, Tree.Kind.CATCH);
                if (ct != null && elt.equals(TreeUtils.elementFromDeclaration(((CatchTree)ct).getParameter())))
                    type.addAnnotation(NONNULL);
            }
            
            annotateIfStatic(elt, type);
            return super.visitIdentifier(node, type);
        }
    }    
}
