package checkers.types;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.Element;
import javax.lang.model.type.TypeKind;

import checkers.types.AnnotatedTypeMirror.*;
import checkers.util.AnnotationUtils;

/**
 * Class to test {@link AnnotatedTypeMirror}s subtype relationships.
 * 
 * The implementation uses the regular traditional java subtyping rules
 * namely, assuming {@code qual(A) := qualifier on A} and 
 * {@code |A| = type without qualifier}: 
 * 
 * <ul>
 *  <li> A <: B  iff |A| <: |B| && qual(A) <: qual(B)</li>
 *  <li> A[] <: B[] iff A <: B and qual(A[]) <: qual(B[])</li>
 *  <li> A<A_1, ..., A_n>  <: B<B_1, ..., B_n>
 *      if A <: B && A_1 = B_1 && ... && A_n = B_n</li>
 * </ul>
 * 
 * Subclasses may need to override the methods as specified below,
 * to override this behavior.
 * 
 * Please note that the array subtyping rules depends on a runtime check to
 * guarantee the soundness of the type system.  It is unsafe for static check
 * purposes.
 */
public class TypeRelations {
    private final AnnotationRelations annotationRelations;
    private final Set<Element> visited;

    /**
     * Constructs an instance of {@code TypeRelations} for the type system
     * whose qualifiers represented in annotationRelations.
     * @param annotationRelations
     */
    public TypeRelations(AnnotationRelations annotationRelations) {
        this.annotationRelations = annotationRelations;
        this.visited = new HashSet<Element>();
    }

    /**
     * Compares two types that are are found within type arguments.
     * 
     * JLS species that for generic types subtype relations, their type arguments
     * need to be equivalent.  This method checks that the types, lhs and rhs,
     * are similar, and have the same annotations found in the current type
     * hierarchy.
     * 
     * Subclasses may override this behavior to implement special semantics.
     * 
     * @param lhs   a type argument
     * @param rhs   a type argument
     * @return  true if the types have the same annotations
     */
    protected boolean compareTypeArg(AnnotatedTypeMirror lhs, AnnotatedTypeMirror rhs) {
        if (lhs.getKind() == TypeKind.TYPEVAR && rhs.getKind() != TypeKind.TYPEVAR) {
            if (visited.contains(lhs.getElement()))
                return true;
            visited.add(lhs.getElement());
            lhs = ((AnnotatedTypeVariable)lhs).getUpperBound();
            if (lhs.getAnnotations().isEmpty())
                return true;
            return isSubtypeImpl(lhs, rhs);
        }
        if (lhs.getKind() == TypeKind.WILDCARD && rhs.getKind() != TypeKind.WILDCARD) {
            if (visited.contains(lhs.getElement()))
                return true;
            visited.add(lhs.getElement());
            lhs = ((AnnotatedWildcardType)lhs).getExtendsBound();
            if (lhs == null) return true;
            return isSubtypeImpl(lhs, rhs);
        }

        List<AnnotationMirror> las = annotationRelations.getAnnotations(lhs.getAnnotations());
        List<AnnotationMirror> ras = annotationRelations.getAnnotations(rhs.getAnnotations());

        boolean result = las.size() == ras.size();
        if (las.size() == 1 && ras.size() == 1) {
            AnnotationMirror la = las.get(0);
            AnnotationMirror ra = ras.get(0);
            if (la == ra)
                result = true;
            result &= la.getAnnotationType().asElement().equals(ra.getAnnotationType().asElement());
            // TODO workaround, until we have a better getElementValuesWithDefaults
            result &= AnnotationUtils.getElementValuesWithDefaults(la)
            .equals(AnnotationUtils.getElementValuesWithDefaults(ra));
        } else if (result) {
            // Iterate and make sure that the lists are equivalent
            Set<Element> laElt = new HashSet<Element>();
            for (AnnotationMirror la : las)
                laElt.add(la.getAnnotationType().asElement());
            Set<Element> raElt = new HashSet<Element>();
            for (AnnotationMirror ra : ras)
                raElt.add(ra.getAnnotationType().asElement());
            result = raElt.equals(raElt);
        }
        
        if (lhs.getKind() == TypeKind.DECLARED && rhs.getKind() == TypeKind.DECLARED)
            result &= compareTypeArguments((AnnotatedDeclaredType)lhs, (AnnotatedDeclaredType)rhs);
        else if (lhs.getKind() == TypeKind.ARRAY && rhs.getKind() == TypeKind.ARRAY)
            result &= compareArrayComponent((AnnotatedArrayType)lhs, (AnnotatedArrayType)rhs);
        return result;
    }

    /**
     * Only compares the type arguments of rhs and lhs.  It returns true if any of
     * them is not a generic type (e.g. one is erased or not a generic class).
     * 
     * JLS specifies that X<A_1, A_2, ..., A_n> is a subtype of 
     * Y<B_1, B_2, ..., B_n> only if A_1 = B_1, ...., A_n = B_n.  Subclasses
     * my override this method to override this behavior.
     * 
     * As an implementation details, the method uses compareTypeArg to compare
     * each type argument individually.
     * 
     * @param lhs   a declared type
     * @param rhs   a declared type
     * @return  true iff the type arguments of lhs and rhs are similar.
     */
    protected boolean compareTypeArguments(AnnotatedDeclaredType lhs, AnnotatedDeclaredType rhs) {
        if (!lhs.getTypeArguments().isEmpty() && !rhs.getTypeArguments().isEmpty()) {
            boolean result = true;
            for (int i = 0; i < lhs.getTypeArguments().size(); ++i) {
                result &= compareTypeArg(lhs.getTypeArguments().get(i), rhs.getTypeArguments().get(i));
            }
            return result;
        }
        return true;
    }

    /**
     * Checks that the rhs component type is a subtype of the lhs component type.
     * 
     * Java specifies that A[] <: B[] if and only if A <: B.  However, given
     * this property makes the code unsafe at run-time, subclasses may override
     * this method to override this behavior.
     * 
     * @param lhs   an array type: left side of an assignment
     * @param rhs   an array type: right side of an assignment
     * @return true iff rhs component type is a subtype of the lhs component type
     */
    protected boolean compareArrayComponent(AnnotatedArrayType lhs, AnnotatedArrayType rhs) {
        AnnotatedTypeMirror rhsComp = rhs.getComponentType();
        AnnotatedTypeMirror lhsComp = lhs.getComponentType();
        return isSubtypeImpl(lhsComp, rhsComp);
    }

    /**
     * Checks whether rhs is a subtype of lhs 
     * 
     * @param lhs   an annotated type
     * @param rhs   an annotated type
     * @return  a true iff rhs a subtype of lhs
     */
    public final boolean isSubtype(AnnotatedTypeMirror lhs, AnnotatedTypeMirror rhs) {
        boolean result = isSubtypeImpl(lhs, rhs);
        this.visited.clear();
        return result;
    }
    
    /**
     * Entry point for subtype checking:
     * Checks if the rhs is a subtype of the lhs according to the annoRelations
     * and the rules, specified above.
     */
    private final boolean isSubtypeImpl(AnnotatedTypeMirror lhs, AnnotatedTypeMirror rhs) {
        if (visited.contains(lhs.getElement()))
            return true;
        if (lhs.getKind() == TypeKind.TYPEVAR && rhs.getKind() != TypeKind.TYPEVAR) {
            lhs = ((AnnotatedTypeVariable)lhs).getUpperBound();
            if (lhs.getAnnotations().isEmpty())
                return true;
            visited.add(lhs.getElement());
        }
        if (lhs.getKind() == TypeKind.WILDCARD && rhs.getKind() != TypeKind.WILDCARD) {
            lhs = ((AnnotatedWildcardType)lhs).getExtendsBound();
            if (lhs == null) return true;
            visited.add(lhs.getElement());
        }
        if (lhs.getKind().isPrimitive() && rhs.getKind() == TypeKind.DECLARED)
            rhs = rhs.typeFactory.getUnboxedType((AnnotatedDeclaredType)rhs);
        if (lhs.getKind() == TypeKind.DECLARED && rhs.getKind().isPrimitive())
            rhs = rhs.typeFactory.getBoxedType((AnnotatedPrimitiveType)rhs);

        AnnotatedTypeMirror rhsBase = rhs.typeFactory.atypes.asSuper(rhs, lhs);
        if (rhsBase != null)
            rhs = rhsBase;
        
        List<AnnotationMirror> lhsAnnotation = annotationRelations.getAnnotations(lhs.getAnnotations());
        List<AnnotationMirror> rhsAnnotation = annotationRelations.getAnnotations(rhs.getAnnotations());
        
        if (lhsAnnotation.isEmpty()) lhsAnnotation.add(null);
        if (rhsAnnotation.isEmpty()) rhsAnnotation.add(null);
        
        boolean result = false;
        for (AnnotationMirror la : lhsAnnotation)
            for (AnnotationMirror ra : rhsAnnotation)
                if (!result && annotationRelations.isSubtype(la, ra))
                    result = true;
        
        if (lhs.getKind() == TypeKind.ARRAY && rhs.getKind() == TypeKind.ARRAY) {
            result &= compareArrayComponent((AnnotatedArrayType)lhs, (AnnotatedArrayType)rhs);
        } else if (lhs.getKind() == TypeKind.DECLARED && rhs.getKind() == TypeKind.DECLARED) {
            result &= compareTypeArguments((AnnotatedDeclaredType)lhs, (AnnotatedDeclaredType)rhs);
        }
            
        return result;
    }
}
