package checkers.util;

import java.lang.annotation.Annotation;
import java.util.*;

import javax.annotation.processing.*;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.*;

import checkers.basetype.BaseTypeChecker;
import checkers.flow.Flow;
import checkers.source.SupportedLintOptions;
import checkers.types.*;

import com.sun.source.tree.*;

/**
 * Implements a checker for type qualifier systems that have only two qualified
 * types where one is a subtype of the other and the qualifiers have no special
 * (implicit) behavior.
 * 
 * <p>
 * 
 * The annotation(s) may be specified on the command line, using annotation
 * processor arguments. Two arguments are used:
 * 
 * <ul>
 * <li>{@code -Aqual}: specifies the annotation for the qualified subtype</li>
 * <li>{@code -Anqual}: (optional) specifies the annotation for the qualified
 * supertype, if one is needed</li>
 * </ul>
 * 
 * <p>
 * 
 * Only the annotation for the qualified subtype must be specified. An
 * annotation for the qualified supertype may be specified; it is used to
 * override a subtype default if the user changes the default.
 */
@SupportedSourceVersion(SourceVersion.RELEASE_7)
@SupportedLintOptions( { "flow" })
@SupportedOptions( { "quals" })
public class CustomChecker extends BaseTypeChecker {

    /** Whether flow is enabled if no lint option is provided. */
    protected static boolean FLOW_BY_DEFAULT = true;
    
    /** For creating annotations (e.g., for flow). */
    private AnnotationFactory annoFactory;

    /** Supported type qualifiers. */
    private Set<Class<? extends Annotation>> quals = null;
    
    @Override @SuppressWarnings("unchecked")
    public synchronized void init(ProcessingEnvironment p) {
        super.init(p);
        annoFactory = new AnnotationFactory(p);
    }

    @Override
    public AnnotatedTypeFactory createFactory(CompilationUnitTree root) {
        AnnotationMirror flowQual = null;
        // FIXME for now we only infer if there's exactly one annotation 
        if (this.quals.size() == 1)
            flowQual = annoFactory.fromName(this.quals.iterator().next().getName());
        return new UserTypeFactory(this, root, flowQual);
    }

    @Override @SuppressWarnings("unchecked")
    public Set<Class<? extends Annotation>> getSupportedTypeQualifiers() {
        // Is annotated?
        Set<Class<? extends Annotation>> supported = 
            super.getSupportedTypeQualifiers();
        if (!supported.isEmpty()) {
            quals = supported;
            return supported;
        }
        
        if (quals != null)
            return quals;
        
        String qualNames = env.getOptions().get("quals");
        if (qualNames == null)
            throw new Error("missing required option: -Aquals");
        
        Set<Class<? extends Annotation>> qualSet = 
            new HashSet<Class<? extends Annotation>>();
        for (String qualName : qualNames.split(",")) {
            try {
                qualSet.add((Class<? extends Annotation>)Class.forName(qualName));
            } catch (ClassNotFoundException e) {
                throw new Error(e);
            }
        }
        
        this.quals = qualSet;
        return quals;
    }

    /**
     * Extends {@link AnnotatedTypeFactory} to optionally use flow-sensitive
     * qualifier inference.
     * 
     * @see {@link Flow}
     */
    private class UserTypeFactory extends AnnotatedTypeFactory {

        private final AnnotationMirror flowQual;
        private final Flow flow;
        private final boolean useFlow;
        private final TypeAnnotator typeAnnotator;
        private final TreeAnnotator treeAnnotator;

        /**
         * Creates a type factory for checking the given compilation unit with
         * respect to the given annotation.
         * 
         * @param checker the checker to which this type factory belongs
         * @param root the compilation unit to scan
         * @param flowQual the annotation inferred by flow-sensitive inference
         */
        UserTypeFactory(CustomChecker checker,
                CompilationUnitTree root, AnnotationMirror flowQual) {
            super(checker, root);

            treeAnnotator = new TreeAnnotator(checker);
            typeAnnotator = new TypeAnnotator(checker);
            
            this.flowQual = flowQual;
            this.flow = new Flow(checker, root, flowQual, this);
            
            this.useFlow = checker.getLintOption("flow", FLOW_BY_DEFAULT);

            // Apply flow-sensitive qualifier inference.
            if (useFlow) flow.scan(root, null);
        }

        @Override
        protected void annotateImplicit(Tree tree, AnnotatedTypeMirror type) {
            treeAnnotator.visit(tree, type);
            if (useFlow && flow.test(tree) == Boolean.TRUE) {
                type.clearAnnotations();
                type.addAnnotation(flowQual);
            }
            typeAnnotator.visit(type);
        }

        @Override
        protected void annotateImplicit(Element elt,
                AnnotatedTypeMirror type) {
            typeAnnotator.visit(type);
        }
    }
}
