package checkers.util;

import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.Name;

import checkers.types.AnnotationRelations;

/**
 * Represents the type qualifier hierarchy of a type system.
 * 
 * It uses a an internal graph to represent the hierarchy.
 * 
 * In addition to simply representing qualifiers, this implementation allows
 * for two special qualifiers:
 * <ul>
 *  <li>Root qualifier: a qualifier that is a super qualifier of all other
 *      types in the hierarchy.  One doesn't need to declare all of its
 *      relations.</li>
 *  <li>Wildcard qualifier: a qualifier that matches any other type.  In other
 *      words is a sub and super of qualifiers.</li>
 * </li>
 */
public class GraphAnnotationRelations extends AnnotationRelations {
    /** map: qualifier --> supertypes of the qualifier**/
    private Map<Name, List<Name>> subtypes;
    
    /** the root of all the qualifiers **/
    private AnnotationMirror root;

    /** the bottom of all qualifiers **/
    private AnnotationMirror wildcard;
    private boolean hasWildcard = false;
    
    /**
     * Constructs an instance of {@code GraphAnnotationRelations} with 
     * no qualifier (i.e. {@code null}) being the root qualifier
     */
    public GraphAnnotationRelations() {
        this(null);
    }
    
    /**
     * Initialize an instance with the provided annotation being the root
     * annotation in the hierarchy
     * 
     * @param root  the type qualifier that is the supertype of all other qualifiers
     */
    public GraphAnnotationRelations(AnnotationMirror root) {
        subtypes = new HashMap<Name, List<Name>>();
        setRootAnnotation(root);
    }
    
    /**
     * Add a type qualifier to the type hierarchy
     * 
     * @param anno  a type qualifier
     * @return  true if is added to the hierarchy graph,
     *          false if it's already in the hierarchy graph
     */
    public boolean addTypeQualifier(AnnotationMirror anno) {
        Name name = AnnotationUtils.annotationName(anno);
        if (subtypes.containsKey(name))
            return false;
        subtypes.put(name, new LinkedList<Name>());
        return true;
    }
    
    /**
     * Adds a subtype relationship between two type qualifiers.
     * One does not need to declare subtyping relations to the root or
     * placeholder type qualifiers.
     * 
     * @param subAnno   the sub type qualifier
     * @param superAnno the super type qualifier
     */
    public void addSubtype(AnnotationMirror subAnno, AnnotationMirror superAnno) {
        addTypeQualifier(subAnno);
        addTypeQualifier(superAnno);

        Name superAnnotation = AnnotationUtils.annotationName(superAnno);
        Name subAnnotation = AnnotationUtils.annotationName(subAnno);
        
        if (AnnotationUtils.isSame(subAnno, superAnno))
            return;
        
        subtypes.get(subAnnotation).add(superAnnotation);
        List<Name> supSupers = subtypes.get(superAnnotation);
        
        if (supSupers != null) {
            for (Name supSuper : supSupers)
                subtypes.get(subAnnotation).add(supSuper);
        }
    }
    
    @Override
    public AnnotationMirror getRootAnnotation() {
        return root;
    }

    /**
     * Set the root type qualifier of the represented hierarchy
     * @param root  the root (the ultimate super) type qualifier
     */
    public void setRootAnnotation(AnnotationMirror root) {
        this.root = root;
        addTypeQualifier(root);
    }

    /**
     * Declare a wildcard annotation that matches all annotations
     * @param placeHolder   the place holder annotation
     */
    public void setWildcard(AnnotationMirror placeHolder) {
        this.wildcard = placeHolder;
        addTypeQualifier(placeHolder);
        this.hasWildcard = true;
    }
    
    @Override
    public boolean isSubtype(AnnotationMirror sup, AnnotationMirror sub) {
        if (AnnotationUtils.isSame(sup, this.root) || AnnotationUtils.isSame(sup, sub))
            return true;
        if (hasWildcard && 
                (AnnotationUtils.isSame(sup, this.wildcard) ||
                        AnnotationUtils.isSame(sub, wildcard)))
            return true;
        Name supAnno = AnnotationUtils.annotationName(sup);
        Name subAnno = AnnotationUtils.annotationName(sub);
        
        return subtypes.get(subAnno).contains(supAnno);
    }
    
    @Override
    public Set<Name> getTypeQualifiers() {
        return Collections.unmodifiableSet(this.subtypes.keySet());
    }
    
    @Override
    public boolean isInConflict(AnnotationMirror a1, AnnotationMirror a2) {
        Name a1Name = AnnotationUtils.annotationName(a1);
        Name a2Name = AnnotationUtils.annotationName(a2);
        return subtypes.containsKey(a1Name) && subtypes.containsKey(a2Name);
    }
}
